;; -*- geiser-scheme-implementation: guile -*-

;; enemy.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td enemy)
  #:use-module (td fset)
  #:use-module (td util)
  #:use-module (td tick)
  #:use-module (td hooks)
  #:use-module (td euclidean)
  #:use-module (td drawable)
  #:use-module (td bullet)
  #:use-module (aa-tree)
  #:use-module (chickadee)
  #:use-module (chickadee scripting)
  #:use-module (srfi srfi-1)
  #:use-module (oop goops)
  #:export (<marching>
            <enemy>
            .prev-x
            .prev-y
            .health-tenth
            enemies
            slow
            damage
            launch-slow-infantry
            launch-three-wheeler
            launch-helo
            launch-pt-boat
            launch-destroyer
            launch-tank
            launch-tracked-soldier
            march-enemies))

(define enemies #nil)

(add-hook! new-game-hook
           (lambda ()
             (set! enemies #nil)))

(define-class <marching> (<drawable>)
  (prev-x #:accessor .prev-x #:init-keyword #:prev-x)
  (prev-y #:accessor .prev-y #:init-keyword #:prev-y)
  (facing #:init-value #:west #:accessor .facing #:init-keyword #:facing)
  (waypoints #:init-keyword #:waypoints #:accessor .waypoints)
  (speed #:init-keyword #:speed #:accessor .speed)
  (slow-time-left #:init-value 0
                  #:accessor .slow-time-left
                  #:init-keyword #:slow-time-left))

(define-class <enemy> (<marching>)
  (health #:init-value #nil #:init-keyword #:health)
  (health-tenth #:init-value #nil)
  (max-health #:accessor .max-health #:init-keyword #:max-health))

(define-method (.health (e <enemy>) x)
  (slot-set! e 'health x))

(define-method (.health (e <enemy>))
  (let ((slot-val (slot-ref e 'health)))
    (if (nil? slot-val)
        (begin (.health e (.max-health e))
               (.max-health e))
        slot-val)))

(define-method (.health-tenth (e <enemy>) x)
  (slot-set! e 'health-tenth x))

(define-method (.health-tenth (e <enemy>))
  (let ((slot-val (slot-ref e 'health-tenth)))
    (if (nil? slot-val)
        (begin (.health-tenth e (calculate-tenth e))
               (.health-tenth e))
        slot-val)))

(define-method (calculate-tenth (e <enemy>))
  (quotient (round (* 10 (.health e)))
            (round (.max-health e))))
                                

(define (slow enemy-uid time)
  (catch 'key-not-found
    (lambda ()
      (let* ((enemy (aa-search enemies enemy-uid))
             (new-slow-time (+ (.slow-time-left enemy) time))
             (new-enemy (slot-fset enemy
                                   'slow-time-left
                                   new-slow-time)))
        (set! enemies (aa-insert enemies enemy-uid new-enemy))))
    (lambda (_ _') #nil)))

(define-method (< (l1 <list>) (l2 <list>))
  (cond ((and (not (nil? l1)) (nil? l2)) #f)
        ((and (nil? l1) (nil? l2)) #f)
        ((and (nil? l1) (not (nil? l2))) #t)
        ((< (car l1) (car l2)) #t)
        ((equal? (car l1) (car l2)) (< (cdr l1) (cdr l2)))
        (else #f)))

(define (damage enemy-uid damage-points)
  (catch 'key-not-found
    (lambda ()
      (let* ((enemy (aa-search enemies enemy-uid))
             (new-hp (- (.health enemy) damage-points)))
        (if (<= new-hp 0)
            (set! enemies (aa-delete enemies enemy-uid))
            (let* ((new-enemy (slot-fset enemy 'health new-hp))
                   (tenth-e (.health-tenth new-enemy (calculate-tenth new-enemy))))
              (set! enemies (aa-insert enemies enemy-uid new-enemy))))))
    (lambda (_ _') #nil)))

(define-method (march (m <marching>) ms)
  (let* ((max-pixels (* ms (.speed m)))
         (next-waypoint (if (null? (.waypoints m))
                            (throw 'arrived)
                            (first (.waypoints m))))
         (wp-x (first next-waypoint))
         (wp-y (second next-waypoint))
         (m-x (.x m))
         (m-y (.y m))
         (speed
          (if (< (.slow-time-left m) 1)
              (.speed m)
              (* (.speed m) 0.5))))
    (if (and (= wp-x m-x) (= wp-y m-y))
        (march (slot-fset m 'waypoints (cdr (.waypoints m))) ms)
        (let* ((d-north (max (- wp-y m-y) 0))
               (d-south (max (- m-y wp-y) 0))
               (d-east (max (- wp-x m-x) 0))
               (d-west (max (- m-x wp-x) 0))
               (to-move-v (if (> d-north d-south)
                              (min d-north (* speed ms))
                              (* -1 (min d-south (* speed ms)))))
               (to-move-h (if (> d-east d-west)
                              (min d-east (* speed ms))
                              (* -1 (min d-west (* speed ms))))))
          (clone m
                 ((.prev-x) m-x)
                 ((.prev-y) m-y)
                 ((.x) (+ to-move-h m-x))
                 ((.y) (+ to-move-v m-y))
                 ((.slow-time-left) (max 0
                                         (- (.slow-time-left m)
                                            period)))
                 ((.facing)
                  (cond
                   ;; top-right
                   ((and (>= to-move-v 0) (>= to-move-h 0))
                    (if (< to-move-v to-move-h)
                        #:east
                        #:north))
                   ;; bottom-right
                   ((and (< to-move-v 0) (>= to-move-h 0))
                    (if (< (abs to-move-v) to-move-h)
                        #:east
                        #:south))
                   ;; top-left
                   ((and (>= to-move-v 0) (< to-move-h 0))
                    (if (< to-move-v (abs to-move-h))
                        #:west
                        #:north))
                   ;; bottom-left
                   (else
                    (if (< (abs to-move-v) (abs to-move-h))
                        #:west
                        #:south)))))))))

(define (launch-slow-infantry path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <enemy>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:texture-proc (lambda (self)
                                 (let ((r (remainder (agenda-time) 1000)))
                                   (cond ((< r 333)
                                          (case (.facing self)
                                            ((#:west) #:soldier-w0)
                                            ((#:north) #:soldier-n0)
                                            ((#:south) #:soldier-s0)
                                            ((#:east) #:soldier-e0)))
                                         ((< r 666)
                                          (case (.facing self)
                                            ((#:west) #:soldier-w1)
                                            ((#:north) #:soldier-n1)
                                            ((#:south) #:soldier-s1)
                                            ((#:east) #:soldier-e1)))
                                         (else
                                          (case (.facing self)
                                            ((#:west) #:soldier-w2)
                                            ((#:north) #:soldier-n2)
                                            ((#:south) #:soldier-s2)
                                            ((#:east) #:soldier-e2))))))
                 #:waypoints path
                 #:speed 0.02
                 #:max-health 6))))

(define* (launch-three-wheeler path #:optional speed)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <enemy>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:texture-proc (lambda (self)
                                 (case (.facing self)
                                   ((#:west) #:three-wheeler-west)
                                   ((#:north) #:three-wheeler-north)
                                   ((#:south) #:three-wheeler-south)
                                   ((#:east) #:three-wheeler-east)))
                 #:waypoints path
                 #:speed (if speed speed 0.1)
                 #:max-health 8))))

(define* (launch-helo path #:optional speed)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <enemy>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:texture-proc (lambda (self)
                                 (case (.facing self)
                                   ((#:west) #:helo-west)
                                   ((#:north) #:helo-north)
                                   ((#:south) #:helo-south)
                                   ((#:east) #:helo-east)))
                 #:waypoints path
                 #:speed (if speed speed 0.2)
                 #:max-health 10))))

(define* (launch-pt-boat path #:optional speed)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <enemy>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:texture-proc (lambda (self)
                                 (case (.facing self)
                                   ((#:west) #:pt-boat-west)
                                   ((#:north) #:pt-boat-north)
                                   ((#:south) #:pt-boat-south)
                                   ((#:east) #:pt-boat-east)))
                 #:waypoints path
                 #:speed (if speed speed 0.1)
                 #:max-health 10))))

(define* (launch-destroyer path #:optional speed)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <enemy>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:texture-proc (lambda (self)
                                 (case (.facing self)
                                   ((#:west) #:destroyer-west)
                                   ((#:north) #:destroyer-north)
                                   ((#:south) #:destroyer-south)
                                   ((#:east) #:destroyer-east)))
                 #:waypoints path
                 #:speed (if speed speed 0.02)
                 #:max-health 100))))

(define* (launch-tank path #:optional speed)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <enemy>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:texture-proc (lambda (self)
                                 (case (.facing self)
                                   ((#:west) #:tank-west)
                                   ((#:north) #:tank-north)
                                   ((#:south) #:tank-south)
                                   ((#:east) #:tank-east)))
                 #:waypoints path
                 #:speed (if speed speed 0.04)
                 #:max-health 30))))

(define* (launch-tracked-soldier path #:optional speed)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <enemy>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:texture-proc (lambda (self)
                                 (case (.facing self)
                                   ((#:west) #:tracked-soldier-west)
                                   ((#:north) #:tracked-soldier-north)
                                   ((#:south) #:tracked-soldier-south)
                                   ((#:east) #:tracked-soldier-east)))
                 #:waypoints path
                 #:speed (if speed speed 0.04)
                 #:max-health 20))))

(define (march-enemies)
  (set! enemies
    (let ((new-tree #nil))
      (aa-for-each (lambda (enemy-pair)
                     (let ((uid (car enemy-pair))
                           (enemy (cdr enemy-pair)))
                       (catch 'arrived
                         (lambda () (set! new-tree
                                      (aa-insert new-tree
                                                 uid
                                                 (march enemy period))))
                         (lambda _
                           (run-hook field-crossed)
                           #nil))))
                   enemies)
      new-tree)))

(add-hook! update-hook (lambda (dt)
                         (unless (game-paused?)
                           (march-enemies))))
