;; -*- geiser-scheme-implementation: guile -*-

;; interface.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td interface)
  #:use-module (srfi srfi-1)
  #:use-module (aa-tree)
  #:use-module (td tower)
  #:use-module (td credits)
  #:use-module (td hooks)
  #:use-module (td gfx)
  #:use-module (td enemy)
  #:use-module (td euclidean)
  #:use-module (td drawable)
  #:use-module (td bullet)
  #:use-module (td tick)
  #:use-module (td grid)
  #:use-module (td level)
  #:use-module (oop goops)
  #:use-module (chickadee)
  #:use-module (chickadee math vector)
  #:use-module (chickadee render font)
  #:use-module (chickadee render sprite)
  #:use-module (chickadee render shapes)
  #:use-module (chickadee render color))

(define fps-label-position (vec2 960.0 728.0))
(define fps-digits-position (vec2 990.0 728.0))
(define new-game-btn-position (vec2 40 728.0))
(define quit-btn-position (vec2 140.0 728.0))
(define basic-turret-btn-position (vec2 300 728.0))
(define resource-turret-btn-position (vec2 380 728.0))
(define delay-turret-btn-position (vec2 460.0 728.0))
(define basic-turret-price-label-position (vec2 300.0 708.0))
(define resource-turret-price-label-position (vec2 380.0 708.0))
(define delay-turret-price-label-position (vec2 460.0 708.0))
(define current-ptr-position-x 0)
(define current-ptr-position-y 0)
(define current-grid-coords (grid-coords 64
                                           current-ptr-position-x
                                           current-ptr-position-y))
(define current-center-coords (center-coords 64
                                             current-ptr-position-x
                                             current-ptr-position-y))
(define current-center-coords-vec (vec2 (first current-center-coords)
                                        (second current-center-coords)))

(define (draw-buttons alpha)
  (draw-text font-gns "NEW GAME" new-game-btn-position)
  (draw-text font-gns "QUIT" quit-btn-position)
  (draw-tile #:basic-turret basic-turret-btn-position)
  (draw-text font-gns "100" basic-turret-price-label-position)
  (draw-tile #:resource-turret resource-turret-btn-position)
  (draw-text font-gns "100" resource-turret-price-label-position)
  (draw-tile #:delay-turret delay-turret-btn-position)
  (draw-text font-gns "100" delay-turret-price-label-position))

(define i-vec-r1 (vec2 0 0)) ;; to avoid allocations
(define i-vec-r2 (vec2 0 0)) ;; to avoid allocations

(define (interpolate x1 y1 x2 y2 alpha)
  (cons
   (+ x1 (* (- x2 x1) alpha))
   (+ y1 (* (- y2 y1) alpha))))

(define (draw-routines alpha)
  (draw-field alpha)
  (draw-buttons alpha)
  (for-each (lambda (t)
              (with-batched-sprites
               (aa-for-each (lambda (p)
                              (let* ((drawable (cdr p))
                                     (interp
                                      (if (and (slot-exists? drawable 'prev-x)
                                               (not (game-paused?)))
                                          (interpolate (.prev-x drawable)
                                                       (.prev-y drawable)
                                                       (.x drawable)
                                                       (.y drawable)
                                                       alpha)
                                          (cons (.x drawable) (.y drawable)))))
                                (set-vec2-x! i-vec-r1 (car interp))
                                (set-vec2-y! i-vec-r1 (cdr interp))
                                (draw-tile (.texture drawable)
                                           i-vec-r1)))
                            t)))
            (list bullets
                  towers
                  enemies))
  (aa-for-each (lambda (p)
                 (let* ((enemy (cdr p))
                        ;; health-x could be stored in enemy to change
                        ;; only after damage
                        (health-x (.health-tenth enemy))
                        (x1 (- (.x enemy) 4))
                        (y1 (+ (.y enemy) 16))
                        (x2 (+ x1 health-x))
                        (y2 y1)
                        (v1 (vec2 x1 y1))
                        (v2 (vec2 x2 y2)))
                   (set-vec2-x! i-vec-r1 x1)
                   (set-vec2-y! i-vec-r1 y1)
                   (set-vec2-x! i-vec-r2 x2)
                   (set-vec2-y! i-vec-r2 y2)
                   (draw-line i-vec-r1 i-vec-r2 #:thickness 1 #:color white))) enemies)
  (draw-credits alpha)
  (draw-crossed 520.0 708.0)
  (draw-frame-rate fps-label-position fps-digits-position)
  (unless (failed-game?)
    (let ((grid-state (catch 'empty-block
                        (lambda ()
                          (get-from-tower-selection-grid current-grid-coords))
                        (lambda (_) #nil))))
      (if (nil? grid-state)
          (draw-tower-hint tower-selection current-center-coords-vec)
          (if (not (or (eq? grid-state #:blk)
                       (eq? grid-state #:basic-turrent-l3)
                       (eq? grid-state #:delay-turret-l3)
                       (eq? grid-state #:resource-turret-l3)))
              (draw-upgrade-hint current-center-coords-vec)))))
    (when (failed-game?)
    (draw-tile #:defeat center-point))
  (when game-complete
    (draw-tile #:victory center-point)))

(define (mouse-press-routines button clicks x y)
  (let* ((g (grid-coords 64 x y))
         (c (center-coords 64 x y))
         (grid-state (catch 'empty-block
                       (lambda ()
                         (get-from-tower-selection-grid g))
                       (lambda (_) #nil))))
    (cond ((and (>= x 20) (< x 110) (<= y 758) (> y 708))
           (run-hook new-game-hook))
          ((and (>= x 120) (< x 190) (<= y 758) (> y 708))
           (abort-game))
          ((and (>= x 270) (< x 330) (<= y 758) (> y 698))
           (set! tower-selection #:basic-turret))
          ((and (>= x 350) (< x 410) (<= y 758) (> y 698))
           (set! tower-selection #:resource-turret))
          ((and (>= x 430) (< x 490) (<= y 758) (> y 698))
           (set! tower-selection #:delay-turret))
          ((> (second g) 9) #nil)
          (else
           (catch 'insufficient-credits
             (lambda ()
               (case grid-state
                 ((#nil)
                  (case tower-selection
                    ((#:basic-turret)
                     (try-purchase! basic-turret-cost)
                     (let ((tower-uid (place-basic-turret (first c) (second c))))
                       (when tower-uid
                         (set! tower-selection-grid (aa-insert tower-selection-grid g #:basic-turret))
                         (set-tower-in-uid-grid! g tower-uid))))
                    ((#:resource-turret)
                     (try-purchase! resource-turret-cost)
                     (let ((tower-uid (place-resource-turret (first c) (second c))))
                       (when tower-uid
                         (set! tower-selection-grid (aa-insert tower-selection-grid g #:resource-turret))
                         (set-tower-in-uid-grid! g tower-uid))))
                    ((#:delay-turret)
                     (try-purchase! delay-turret-cost)
                     (let ((tower-uid (place-delay-turret (first c) (second c))))
                       (when tower-uid
                         (set! tower-selection-grid (aa-insert tower-selection-grid g #:delay-turret))
                         (set-tower-in-uid-grid! g tower-uid))))
                    (else #nil)))
                 ((#:basic-turret)
                  (try-purchase! basic-turret-l2-cost)
                  (let ((tower-uid (place-basic-turret-l2 (first c) (second c))))
                    (set! tower-selection-grid (aa-insert tower-selection-grid g #:basic-turret-l2))
                    (delete-tower-from-towers! (get-uid-from-tower-uid-grid g))
                    (set-tower-in-uid-grid! g tower-uid)))
                 ((#:basic-turret-l2)
                  (try-purchase! basic-turret-l3-cost)
                  (let ((tower-uid (place-basic-turret-l3 (first c) (second c))))
                    (set! tower-selection-grid (aa-insert tower-selection-grid g #:basic-turret-l3))
                    (delete-tower-from-towers! (get-uid-from-tower-uid-grid g))
                    (set-tower-in-uid-grid! g tower-uid)))
                 ((#:resource-turret)
                  (try-purchase! resource-turret-l2-cost)
                  (let ((tower-uid (place-resource-turret-l2 (first c) (second c))))
                    (set! tower-selection-grid (aa-insert tower-selection-grid g #:resource-turret-l2))
                    (delete-tower-from-towers! (get-uid-from-tower-uid-grid g))
                    (set-tower-in-uid-grid! g tower-uid)))
                 ((#:resource-turret-l2)
                  (try-purchase! resource-turret-l3-cost)
                  (let ((tower-uid (place-resource-turret-l3 (first c) (second c))))
                    (set! tower-selection-grid (aa-insert tower-selection-grid g #:resource-turret-l3))
                    (delete-tower-from-towers! (get-uid-from-tower-uid-grid g))
                    (set-tower-in-uid-grid! g tower-uid)))
                 ((#:delay-turret)
                  (try-purchase! delay-turret-l2-cost)
                  (let ((tower-uid (place-delay-turret-l2 (first c) (second c))))
                    (set! tower-selection-grid (aa-insert tower-selection-grid g #:delay-turret-l2))
                    (delete-tower-from-towers! (get-uid-from-tower-uid-grid g))
                    (set-tower-in-uid-grid! g tower-uid)))
                 ((#:delay-turret-l2)
                  (try-purchase! delay-turret-l3-cost)
                  (let ((tower-uid (place-delay-turret-l3 (first c) (second c))))
                    (set! tower-selection-grid (aa-insert tower-selection-grid g #:delay-turret-l3))
                    (delete-tower-from-towers! (get-uid-from-tower-uid-grid g))
                    (set-tower-in-uid-grid! g tower-uid)))
                 (else #nil)))
             (lambda (_) #nil))))))

(add-hook! draw-hook
           (lambda (alpha)
             (draw-routines alpha)))

(add-hook! mouse-press-hook
           (lambda (a b c d)
             (mouse-press-routines a b c d)))

(add-hook! mouse-move-hook
           (lambda (x y _ __ ___)
             (set! current-ptr-position-x x)
             (set! current-ptr-position-y (- 768 y))
             (set! current-grid-coords (grid-coords 64
                                                    current-ptr-position-x
                                                    current-ptr-position-y))
             (set! current-center-coords (center-coords 64
                                                        current-ptr-position-x
                                                        current-ptr-position-y))
             (set-vec2-x! current-center-coords-vec
                          (first current-center-coords))
             (set-vec2-y! current-center-coords-vec
                          (second current-center-coords))))

(add-hook! game-fail
           (lambda ()
             (pause-game)))
