;; -*- geiser-scheme-implementation: guile -*-

;; level.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td level)
  #:use-module (chickadee)
  #:use-module (chickadee math vector)
  #:use-module (chickadee scripting)
  #:use-module (chickadee scripting script)
  #:use-module (chickadee render sprite)
  #:use-module (chickadee render texture)
  #:use-module (chickadee render font)
  #:use-module (td hooks)
  #:use-module (td enemy)
  #:use-module (td tower)
  #:use-module (td gfx)
  #:use-module (aa-tree)
  #:export (draw-field
            failed-game?
            draw-crossed
            game-complete))

(define max-to-cross 10)

(define total-crossed 0)
(define current-level-script #nil)
(define game-complete #f)

(define (draw-field alpha)
  (draw-sprite field (vec2 0.0 0.0)))

(define path1 '((992 544) (800 544) (800 416) (864 416) (864 160)
                (672 160) (672 224) (480 224) (480 352) (544 352)
                (544 544) (160 544) (160 416) (288 416) (288 160)
                (160 160) (160 224) (32 224)))

(define path1-river-pre '((4 0) (6 0) (6 6) (9 6) (9 8) (11 8) (11 11)))

(define path1-river '((288 32) (416 32) (416 416) (608 416) (608 544) (736 544) (736 736)))

(define path1-center '((992 352) (32 352)))
(define path1-upper '((992 224) (32 224)))
(define path1-lower '((992 480) (32 480)))

(add-hook! load-hook (lambda () (run-hook new-game-hook)))

(define (failed-game?)
  (>= total-crossed max-to-cross))

(add-hook! field-crossed
           (lambda ()
             (set! total-crossed (1+ total-crossed))
             (if (failed-game?)
                 (run-hook game-fail))))

(add-hook! new-game-hook
           (lambda ()
             (set! total-crossed 0)))

(define (draw-crossed x y)
  (draw-text font-gns
             (string-append "Fails: "
                            (number->string total-crossed)
                            " of "
                            (number->string max-to-cross))
             (vec2 x y)))

(define (blocked-tiles-to-tower-selection-grid cols blk-lst)
  (letrec ((f (lambda (bl col row tg)
                (let ((end-of-row (= (1+ col) cols)))
                  (if (nil? bl)
                      tg
                      (f (cdr bl)
                         (if end-of-row
                             0
                             (1+ col))
                         (if end-of-row
                             (1+ row)
                             row)
                         (if (nil? (car bl))
                             tg
                             (aa-insert tg
                                        (list col row)
                                        (car bl)))))))))
    (f blk-lst 0 0 #nil)))

(define level1-blocked-tiles
  '(#nil #nil #nil #nil :blk :blk :blk #nil #nil #nil #nil #nil #nil #nil #nil #nil
         #nil #nil #nil #nil #nil #nil :blk #nil #nil #nil #nil #nil #nil #nil #nil #nil
         #nil #nil :blk :blk :blk #nil :blk #nil #nil #nil :blk :blk :blk :blk #nil #nil
         :blk :blk :blk #nil :blk #nil :blk :blk :blk :blk :blk #nil #nil :blk #nil #nil
         #nil #nil #nil #nil :blk #nil :blk :blk #nil #nil #nil #nil #nil :blk #nil #nil
         #nil #nil #nil #nil :blk #nil :blk :blk :blk #nil #nil #nil #nil :blk #nil #nil
         #nil #nil :blk :blk :blk #nil :blk :blk :blk :blk #nil #nil :blk :blk #nil #nil
         #nil #nil :blk #nil #nil #nil #nil #nil :blk :blk #nil #nil :blk #nil #nil #nil
         #nil #nil :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk
         #nil #nil #nil #nil #nil #nil #nil #nil #nil #nil #nil :blk #nil #nil #nil #nil
         :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk
         :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk :blk))

(define level1-tower-selection-grid (blocked-tiles-to-tower-selection-grid 16 level1-blocked-tiles))

(add-hook! new-game-hook
           (lambda ()
             (set! tower-selection-grid level1-tower-selection-grid)))

(define (min2ms min)
  (* min 60000))

(define (sec2ms sec)
  (* sec 1000))

(define (spaced-repeat iter ms proc)
  (letrec ((fn
            (lambda (i)
              (unless (< i 1)
                (proc)
                (sleep ms)
                (fn (1- i))))))
    (fn iter)))

(define (sleep-until-no-enemies)
  (while (not (nil? enemies))
    (sleep 1000)))

(define (sleep20sec) (sleep 20000))

(define level1-script
  (lambda ()
    (set! current-level-script
      (script
       (sleep (sec2ms 5))

       ;; WAVE: 1 INFANTRY
       (spaced-repeat 1 (sec2ms 2)
                      (lambda () (launch-slow-infantry path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE 3 INFANTRY
       (spaced-repeat 3 (sec2ms 2)
                      (lambda () (launch-slow-infantry path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 6 INFANTRY
       (spaced-repeat 6 (sec2ms 2)
                      (lambda () (launch-slow-infantry path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 1 THREE WHEELER
       (launch-three-wheeler path1)
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 9 INFANTRY
       (spaced-repeat 9 (sec2ms 2)
                      (lambda () (launch-slow-infantry path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 3 THREE-WHEELERS
       (spaced-repeat 3 (sec2ms 2)
                      (lambda () (launch-three-wheeler path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 2 TANKS
       (spaced-repeat 2 (sec2ms 3)
                      (lambda () (launch-tank path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 12 INFANTRY AND 2 TANKS
       (spaced-repeat 4 (sec2ms 2)
                      (lambda () (launch-slow-infantry path1)))
       (sleep (sec2ms 3))
       (launch-tank path1 0.02)
       (sleep (sec2ms 3))
       (spaced-repeat 4 (sec2ms 2)
                      (lambda () (launch-slow-infantry path1)))
       (sleep (sec2ms 3))
       (launch-tank path1 0.02)
       (sleep (sec2ms 3))
       (spaced-repeat 4 (sec2ms 2)
                      (lambda () (launch-slow-infantry path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 3 PT-BOATS
       (spaced-repeat 3 750
                      (lambda () (launch-pt-boat path1-river)))
       (sleep-until-no-enemies)
       (sleep20sec)       
       
       ;; WAVE: 8 THREE-WHEELERS
       (spaced-repeat 8 750
                      (lambda () (launch-three-wheeler path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 12 TANKS
       (spaced-repeat 12 (sec2ms 3)
                      (lambda () (launch-tank path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 3 DESTROYERS
       (spaced-repeat 3 (sec2ms 6)
                      (lambda () (launch-destroyer path1-river)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 24 TRACKED SOLDIER AND 24 TANKS
       (spaced-repeat 12 100
                      (lambda ()
                        (spaced-repeat 2 1500
                                       (lambda () (launch-tracked-soldier path1)))
                        (spaced-repeat 2 (sec2ms 2)
                                       (lambda () (launch-tank path1)))))
       (sleep-until-no-enemies)
       (sleep20sec)

       (spaced-repeat 6 500
                      (lambda () (launch-helo path1-center)))
       (sleep20sec)
       (sleep-until-no-enemies)

       ;; WAVE: 24 THREE-WHEELERS
       (spaced-repeat 24 (sec2ms 2)
                      (lambda () (launch-three-wheeler path1)))
       (sleep-until-no-enemies)
       (sleep20sec)

       ;; WAVE: 24 TRACKED SOLDIER, 24 TANKS, 6 destroyers, 18 helos
       (spaced-repeat 6 100
                      (lambda ()
                        (launch-tracked-soldier path1)
                        (sleep 1500)
                        (launch-tank path1)
                        (sleep 2000)
                        (launch-tracked-soldier path1)
                        (sleep 1500)
                        (launch-tank path1)
                        (sleep 2000)
                        
                        (launch-destroyer path1-river)
                        
                        (launch-tracked-soldier path1)
                        (sleep 1500)
                        (launch-tank path1)
                        (sleep 2000)
                        (launch-tracked-soldier path1)
                        (sleep 1500)
                        (launch-tank path1)
                        
                        (launch-helo path1-center)
                        (sleep 1000)
                        (launch-helo path1-lower)
                        (sleep 1000)
                        (launch-helo path1-upper)
                        (sleep 1000)))
       (sleep-until-no-enemies)

       ;; **** GAME OVER ****
       (sleep (sec2ms 3))
       (set! game-complete #t)
       (run-hook game-victory-hook)
       ))))

(add-hook! new-game-hook
           (lambda ()
             (set! game-complete #f)
             (unless (nil? current-level-script)
               (cancel-script current-level-script))
             (level1-script)))
