;; -*- geiser-scheme-implementation: guile -*-

;; tower.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; Game configuration parameters

(define-module (td tower)
  #:use-module (td tick)
  #:use-module (aa-tree)
  #:use-module (td hooks)
  #:use-module (td util)
  #:use-module (td bullet)
  #:use-module (oop goops)
  #:use-module (td euclidean)
  #:use-module (td drawable)
  #:use-module (td fset)
  #:use-module (td enemy)
  #:use-module (td credits)
  #:use-module (chickadee)
  #:export (towers
           tower-procedures
           tower-selection
           tower-selection-grid
           tower-uid-grid
           <tower>
           place-basic-turret
           place-basic-turret-l2
           place-basic-turret-l3
           place-delay-turret
           place-delay-turret-l2
           place-delay-turret-l3
           place-resource-turret
           place-resource-turret-l2
           place-resource-turret-l3
           process-tower-procedures
           delete-tower-from-towers!
           set-tower-in-uid-grid!
           get-uid-from-tower-uid-grid
           get-from-tower-selection-grid
           basic-turret-range
           basic-turret-reload-time
           basic-turret-damage
           basic-turret-cost
           basic-turret-l2-range
           basic-turret-l2-reload-time
           basic-turret-l2-damage
           basic-turret-l2-cost
           basic-turret-l3-range
           basic-turret-l3-reload-time
           basic-turret-l3-damage
           basic-turret-l3-cost
           delay-turret-range
           delay-turret-reload-time
           delay-turret-delay-time
           delay-turret-cost
           delay-turret-l2-range
           delay-turret-l2-reload-time
           delay-turret-l2-delay-time
           delay-turret-l2-cost
           delay-turret-l3-range
           delay-turret-l3-reload-time
           delay-turret-l3-delay-time
           delay-turret-l3-delay-time
           delay-turret-l3-damage
           delay-turret-l3-cost
           resource-turret-reload-time
           resource-turret-credit-increase
           resource-turret-cost
           resource-turret-l2-reload-time
           resource-turret-l2-credit-increase
           resource-turret-l2-cost
           resource-turret-l3-reload-time
           resource-turret-l3-credit-increase
           resource-turret-l3-cost))

(define basic-turret-range 92)
(define basic-turret-reload-time 2000)
(define basic-turret-damage 1)
(define basic-turret-cost 100)

(define basic-turret-l2-range 158)
(define basic-turret-l2-reload-time 1500)
(define basic-turret-l2-damage 1)
(define basic-turret-l2-cost 90)

(define basic-turret-l3-range 158)
(define basic-turret-l3-reload-time 1200)
(define basic-turret-l3-damage 1.2)
(define basic-turret-l3-cost 90)

(define delay-turret-range 92)
(define delay-turret-reload-time 2000)
(define delay-turret-delay-time 6000)
(define delay-turret-cost 100)

(define delay-turret-l2-range 158)
(define delay-turret-l2-reload-time 1500)
(define delay-turret-l2-delay-time 7000)
(define delay-turret-l2-cost 90)

(define delay-turret-l3-range 158)
(define delay-turret-l3-reload-time 1200)
(define delay-turret-l3-delay-time 7000)
(define delay-turret-l3-delay-time 8000)
(define delay-turret-l3-damage 0.2)
(define delay-turret-l3-cost 90)

(define resource-turret-reload-time 3000)
(define resource-turret-credit-increase 1)
(define resource-turret-cost 100)

(define resource-turret-l2-reload-time 2250)
(define resource-turret-l2-credit-increase 1)
(define resource-turret-l2-cost 90)

(define resource-turret-l3-reload-time 1688)
(define resource-turret-l3-credit-increase 1)
(define resource-turret-l3-cost 90)

;; Global data structures

(define towers #nil)
(define tower-procedures #nil)
(define tower-selection #:basic-turret)
(define tower-selection-grid #nil)
(define tower-uid-grid #nil)

(define (add-tower! new-tower-id new-tower)
  (set! towers (aa-insert towers
                          new-tower-id
                          new-tower)))

(define (add-tower-procedure! proc)
  (set! tower-procedures (cons proc tower-procedures)))

(define (get-from-tower-selection-grid grid-coords)
  (catch 'key-not-found
    (lambda ()
      (aa-search tower-selection-grid grid-coords))
    (lambda (_ _') (throw 'empty-block))))

(define (get-uid-from-tower-uid-grid grid-coords)
  (catch 'key-not-found
    (lambda ()
      (aa-search tower-uid-grid grid-coords))
    (lambda (_ _') (throw 'empty-block))))

(define (set-tower-in-uid-grid! grid-coords tower-uid)
  (set! tower-uid-grid
    (aa-insert tower-uid-grid
               grid-coords
               tower-uid)))

(define (delete-tower-from-towers! uid)
  (set! towers
    (aa-delete towers uid)))

(add-hook! new-game-hook
           (lambda ()
             (set! towers #nil)
             (set! tower-procedures #nil)
             (set! tower-selection #:basic-turret)))

(define-class <tower> (<drawable>)
  (reload-time-left #:init-value 1000
                    #:accessor .reload-time-left
                    #:init-keyword #:reload-time-left))

(define-method (pass-reload-time tower-id (t <tower>))
  (set! towers (aa-insert towers
                          tower-id
                          (slot-fset t
                                     'reload-time-left
                                     (- (.reload-time-left t)
                                        period)))))

(define-method (reset-reload-time tower-id (t <tower>) time)
  (set! towers (aa-insert towers
                          tower-id
                          (slot-fset t
                                     'reload-time-left
                                     time))))

(define (launching-turret-proc new-tower
                               new-tower-id
                               range
                               reload-time
                               launch-proc)
  (lambda ()
    (catch 'key-not-found
      (lambda ()
        (let ((tower (aa-search towers new-tower-id)))
          (if (> (.reload-time-left tower) 0)
              ;; Still reloading
              (pass-reload-time new-tower-id tower)
              ;; Fire away!
              (catch 'no-candidate
                (lambda ()
                  (let* ((nearest-enemy-id (nearest new-tower
                                                    enemies
                                                    range))
                         (nearest-enemy (aa-search enemies
                                                   nearest-enemy-id))
                         (distance (dist new-tower nearest-enemy)))
                    (launch-proc (.x new-tower)
                                 (.y new-tower)
                                 nearest-enemy-id)
                    (reset-reload-time new-tower-id tower reload-time)))
                ;; Nothing to fire at
                (lambda (_) (pass-reload-time new-tower-id tower))))))
      ;; Tower is gone
      (lambda (_ _') (throw 'terminate)))))

(define (place-basic-turret x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:basic-turret)))) 
    (add-tower! new-tower-id new-tower)
    (add-tower-procedure!
     (launching-turret-proc new-tower
                            new-tower-id
                            basic-turret-range
                            basic-turret-reload-time
                            (lambda (x y nearest-enemy-id)
                              (launch-simple-bullet x
                                                    y
                                                    nearest-enemy-id
                                                    basic-turret-damage))))
    new-tower-id))

(define (place-basic-turret-l2 x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:basic-turret-l2))))
    (add-tower! new-tower-id new-tower)
    (add-tower-procedure!
     (launching-turret-proc new-tower
                            new-tower-id
                            basic-turret-l2-range
                            basic-turret-l2-reload-time
                            (lambda (x y nearest-enemy-id)
                              (launch-simple-bullet x
                                                    y
                                                    nearest-enemy-id
                                                    basic-turret-l2-damage))))
    new-tower-id))

(define (place-basic-turret-l3 x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:basic-turret-l3))))
    (add-tower! new-tower-id new-tower)
    (add-tower-procedure!
     (launching-turret-proc new-tower
                            new-tower-id
                            basic-turret-l3-range
                            basic-turret-l3-reload-time
                            (lambda (x y nearest-enemy-id)
                              (launch-fire-bullet x
                                                  y
                                                  nearest-enemy-id
                                                  basic-turret-l3-damage))))
    new-tower-id))

(define (place-delay-turret x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:delay-turret))))
    (add-tower! new-tower-id new-tower)
    (add-tower-procedure!
     (launching-turret-proc new-tower
                            new-tower-id
                            delay-turret-range
                            delay-turret-reload-time
                            (lambda (x y nearest-enemy-id)
                              (launch-delay-bullet (.x new-tower)
                                                   (.y new-tower)
                                                   delay-turret-delay-time
                                                   nearest-enemy-id))))
    new-tower-id))

(define (place-delay-turret-l2 x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:delay-turret-l2))))
    (add-tower! new-tower-id new-tower)
    (add-tower-procedure!
     (launching-turret-proc new-tower
                            new-tower-id
                            delay-turret-l2-range
                            delay-turret-l2-reload-time
                            (lambda (x y nearest-enemy-id)
                              (launch-delay-bullet (.x new-tower)
                                                   (.y new-tower)
                                                   delay-turret-l2-delay-time
                                                   nearest-enemy-id))))
    new-tower-id))

(define (place-delay-turret-l3 x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:delay-turret-l3))))
    (add-tower! new-tower-id new-tower)
    (add-tower-procedure!
     (launching-turret-proc new-tower
                            new-tower-id
                            delay-turret-l3-range
                            delay-turret-l3-reload-time
                            (lambda (x y nearest-enemy-id)
                              (launch-damaging-delay-bullet (.x new-tower)
                                                            (.y new-tower)
                                                            delay-turret-l3-delay-time
                                                            delay-turret-l3-damage
                                                            nearest-enemy-id))))
    new-tower-id))

(define (resource-turret-proc new-tower
                              new-tower-id
                              reload-time
                              credit-increase)
  (lambda ()
    (catch 'key-not-found
      (lambda ()
        (let ((tower (aa-search towers new-tower-id)))
          (if (> (.reload-time-left tower) 0)
              ;; Still reloading
              (pass-reload-time new-tower-id tower)
              ;; Generate credit
              (begin
                (deposit-credits! credit-increase)
                (reset-reload-time new-tower-id tower reload-time)))))
      ;; Tower is gone
      (lambda (_ _') (throw 'terminate)))))

(define (place-resource-turret x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:resource-turret))))
    (set! towers (aa-insert towers
                            new-tower-id
                            new-tower))

    (add-tower-procedure!
     (resource-turret-proc new-tower
                           new-tower-id
                           resource-turret-reload-time
                           resource-turret-credit-increase))
    new-tower-id))

(define (place-resource-turret-l2 x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:resource-turret-l2))))
    (set! towers (aa-insert towers
                            new-tower-id
                            new-tower))

    (add-tower-procedure!
     (resource-turret-proc new-tower
                           new-tower-id
                           resource-turret-l2-reload-time
                           resource-turret-l2-credit-increase))
    new-tower-id))

(define (place-resource-turret-l3 x y)
  (let ((new-tower-id (get-uid))
        (new-tower (make <tower>
                     #:x x #:y y
                     #:texture-proc (lambda (self) #:resource-turret-l3))))
    (set! towers (aa-insert towers
                            new-tower-id
                            new-tower))

    (add-tower-procedure!
     (resource-turret-proc new-tower
                           new-tower-id
                           resource-turret-l3-reload-time
                           resource-turret-l3-credit-increase))
    new-tower-id))

(define (process-tower-procedures)
  (let ((new-queue #nil))
    (for-each (lambda (p)
                (catch 'terminate
                  (lambda ()
                    (p)
                    (set! new-queue (cons p new-queue)))
                  (lambda (_) #nil)))
              tower-procedures)
    (set! tower-procedures new-queue)))

(add-hook! update-hook
           (lambda (dt)
             (unless (game-paused?)
               (process-tower-procedures))))
